/*****************************************************************************
Author:Peter J. Vis
First Written:8 Dec 2004
Last Updated: 12 Dec 2006
Title: Displaying text on a 128x64 LCD

Microcontroller:ATmega32
Crystal:16Mhz
Platform:Development System

URL:http://petervis.co.cc/C/128x64/128x64.html

LIMITATIONS:
No portion of this work can be used in commercial
applications without prior written permission.This program can be published in
magazines and articles with previous consent and as long as credit is given
to the author.

COPYRIGHT:	
All the material including function names are copyright.

LIABILITY:
Obviously none.

SUPPORT:
Feel free to contact me if you are developing a product and wish to employ my
services. 

POSTCARD:
Send me a postcard / email if you found my code useful. 

PURPOSE:
This program and its associated files 128x64.c and 128x64.h were created to
show how the LCD could be used to display a bitmapped graphic.

CIRCUIT:
My Modular Development System

*******************************************************************************/


// mine
#include "128x64-text.h"

// The font file is generated by my font extractor program written
// in Visual Basic
#include "font_data.h" 

// built-in
#include <inttypes.h>
#include <util/delay.h>








/* ---------------------------------------------------------------
	Initialise the LCD according to the DATASHEET. I prefer to 
	go through a reset procedures.
	
	clearl will clear all the 128x64 dots. and reset the column
	and page to 0. In noisy environments, this function provides
	a further insurance of starting with a clean screen.
 -----------------------------------------------------------------*/
void lcdInit (void)
{  
	// set the data and control ports for output.
	// and reset them to 0	
	cPort_ddr = 0xff; 
	cPort = 0x00; 

	dPort_ddr = 0xff;
	dPort = 0x00;  

	// reset the lcd using it's reset pin 17
	cPort |= (1<<lcdrst); 
  
	 
	WriteCommand(DISP_OFF);	  // according to the datasheet
	WriteCommand(START_LINE);	
	WriteCommand(DISP_ON);  
  
	// clear every pixel and reset the column and page
	// to 0
	clearl(); 
}





/* ---------------------------------------------------------------
	This will set to the required page 0-7, and reset the column 
	back to starting point, column 0. 
 -----------------------------------------------------------------*/
void setNewPage (unsigned char PageData)
{		
		PageData = PageData + SET_PAGE;
        WriteCommand(PageData);	
		
		WriteCommand(SET_ADDRESS); // reset column address to first
}								   // column again




/* ---------------------------------------------------------------
	This is to clear the LCD screen, and also resets the column 
	and page registers back to the beginning.
	
	Use this function when the whole screen needs clearing to 
	display another text / graphic.
 -----------------------------------------------------------------*/
void clearl (void){
 
 unsigned int i;
 
 // reset global variables
 column = 0;							// reset column and page to beginning
 page = 0;
 
 // reset LCD registers
  
 WriteCommand(SET_ADDRESS);				// reset column and page to the beginning
 WriteCommand(SET_PAGE);
 
 
 for(i = 0; i < 1024 ;i++)
 {
  
  if (column == 128)
  {
   column = 0;
   page++;
   setNewPage(page);
  }

  cSelect(column);						
										
	DataMode(); 						
	dPort = 0b00000000; 				// clear the pixels
	LocknLoad();						

  column++;     						// increase column (maximum 128).
 }
 
 // reset global variables
 column = 0;							// reset column and page to beginning
 page = 0;
 
 // reset LCD registers

 WriteCommand(SET_ADDRESS);				// reset column and page to the beginning
 WriteCommand(SET_PAGE);
 
}





 
 /* ---------------------------------------------------------------
	set LCD to data mode to accepts pixel
	data through the data port.
 -----------------------------------------------------------------*/
void DataMode (void)
{   
  //set pins for data mode
  cPort |= (1<<lcdrs);
  cPort &= ~(1<<lcdrw); 
}






/* ---------------------------------------------------------------
	to send commands to the LCD Controllers through the data port.
	
	** use this BEFORE cSelect() otherwise previous settings will 
	be lost. This function sends command to both SC1 & CS2
 -----------------------------------------------------------------*/
void WriteCommand(uint8_t cmd) {

		cPort |= (1<<lcdcs1);   //Enable CS1 & CS2
        cPort |= (1<<lcdcs2);   
        		
        cPort &= ~(1<<lcdrw);   //Clear RW & RS for Command mode   
		cPort &= ~(1<<lcdrs);   
		
    dPort = cmd;   // write command
    LocknLoad();
}






/* ---------------------------------------------------------------
	This selects between the controller chip CS1 / CS2
	based on the current column number. Select which half of the 
	LCD to send the data to.
 -----------------------------------------------------------------*/
void cSelect(uint8_t col)
{

if (col <= 63){
	cPort |= (1<<lcdcs1);   //Enable CS1 
	cPort &= ~(1<<lcdcs2);  //Disable CS2 	
	}
	
	if (col >= 64){
	cPort |= (1<<lcdcs2);   //Enable CS2 
	cPort &= ~(1<<lcdcs1);  //Disable CS1 	
	}

}


/* ---------------------------------------------------------------
	According to the data sheet the enable pin has to be cleared
	for the data port to be read! Set it to 1 to lock the data
	then clear it to load the data. This lock and load cycle is
	1us (min) in ideal conditions, however realistically I am 
	using 3us because of circuit noise etc... to give a wide 
	margin of error thereby offering stability in a wide range of
	conditions and applications.
	
	Enable pin e is set to 1 to output data and back to 0 to 
	latch data.
				  ____________
        _________|			  |______________    Data DB0-DB7
	             |____________|
				 				 
	             ______________               __ Enable pin e
                |              |             |
        ________|              |_____________|

                |    0.45us    |    0.45us   |
                |                            |
                |<----        1us       ---->|      
 -----------------------------------------------------------------*/
void LocknLoad (void)
{
    cPort |= (1<<lcde);   //Output - Data Load
	//delay
   _delay_us(4);
   
    cPort &= ~(1<<lcde);  //Latch - Data Lock
	//delay
   _delay_us(4);
}





/* ---------------------------------------------------------------
	This is to set the page and column registers of the LCD based
	on the column and row values given.
	
	Use this prior to printing text on the LCD, when the text is
	needed at a particulat point x,y on the screen.
	
	GLOBALS:page and column
 -----------------------------------------------------------------*/
void setXY(uint8_t x, uint8_t y) {

	uint8_t cmd;
	
	if(x > 127) x = 0;				// validate coordinates
	if(y > 63)  y = 0;								
	
	page = y/8;
	column = x;
	cSelect(column); 				// select the chip based on the column number
									// column is a global variable it's contents are used
									// by the ascii driver lcd_putch()
	
	
	cmd = SET_ADDRESS | x;	
	WriteCommand(cmd);
	
	cmd = SET_PAGE | page;			// set page, or y address on the selected chip
	WriteCommand(cmd);

}






/* ---------------------------------------------------------------
	My ASCII character Driver:

	Reads the character bitmap array font header files and prints
	it to the LCD.
	
	Has End of line feature, occurs when a whole character cannot
	fit at the end of the line.
	
	use setXY() to first set the position where you want to print
	
	page and column variables are set by setXY() and 
	preserved as global, for subsequent character printing calls.
	
	LIMITATION: avoid printing a character overlapping the 64 col,
	however space there is ok. most fonts normally manage to avoid 
	that line.
 -----------------------------------------------------------------*/
void lcd_writech(uint8_t ch)  {

		

			// make sure the characters to print are within the ascii range values
			// as specified in the font bitmap file
			// if not they are modified to the outermost limits
			if (ch < ASCIIL) ch = ASCIIL;  	
			if (ch > ASCIIH) ch = ASCIIH;

			
			int base;	//HOLDING A CHAR VALUE **THIS MUST BE A INT AND NOT uint8_t
			uint8_t i, p, cmd;
  
			uint8_t tmppage=page;							// temporary page counter
			uint8_t tmpcol=column;							// temporary column counter for one character
			uint8_t EOL;	

			EOL = column + CHAR_FW;	
			base = (ch - ASCIIL) * CHAR_PG * CHAR_FW; 		// array starts from 0 but ascii codes start from ASCIIL
															// low range (specified by the user)
															// CHAR_FW, character width in bytes.
															// CHAR_PG, number of pages the character occupies.
															// Base is the first byte of the character bitmap
															// in the array.
								
			// check the column position space to make sure it can fit the whole character
			// if  a whole character cannot be placed, then End Of Line...		
			if (EOL > 128){	
			
				column = 0;
				tmpcol = 0;
				cSelect(column); 							// selects the first half of the
															// lcd, column 0
				WriteCommand(SET_ADDRESS); 					// reset column address to first column in LCD as well	
			
				page = page + CHAR_PG;						// next line, number of pages down
				tmppage=page;
				cmd = SET_PAGE | page;						
				WriteCommand(cmd);
				
			}
								
			for (p=0;p<CHAR_PG;p++){	
			
				for(i = 0; i < CHAR_FW ;i++){ 				// CHAR_FW number of bytes for each character	
													
	
				cSelect(tmpcol);	
					DataMode(); 							// prepare mode to send a byte of data
					dPort = pgm_read_byte(&basicfont[base++]); 	// send data from pgm space to the data port				
					LocknLoad();							// load data and relatch
				
					tmpcol++;
				}

				if(CHAR_PG>1){								// if a  multipage character
				   
				  tmppage = tmppage + 1;

				  cmd = SET_PAGE | tmppage;					// set next page for the multi-page character
				  WriteCommand(cmd);				  
				
				  cmd = SET_ADDRESS | column;				// reset the column back to the starting point
				  WriteCommand(cmd);
				  tmpcol = column;	
				  
				}

			}
			
					
							
			// only put a space if it can fit before the end of line
			// otherwise you dont need a space at the end.
			// also column placement for the next character is set here
			if ((column + CHAR_SPACE) < 128){											
				column = column + CHAR_FW + CHAR_SPACE;		// next character position
				cmd = SET_ADDRESS | column;				    // reset the column to the next character position
				WriteCommand(cmd);												  
			}
						
		
			if(CHAR_PG>1){									// if a  multipage character				  
				cmd = SET_PAGE | page;						// return back to the starting page
				WriteCommand(cmd);			  				  
			}
			
				
}


				



/* ---------------------------------------------------------------
	This is the command shell. Use this in the main body of the 
	program to print text to the LCD.
	
	e.g. printl("hello world");
 -----------------------------------------------------------------*/
void printl(const char* str) {
  while(*str) {
    lcd_writech(*str++);
  }
}




